/* FXS2OP.C - mixed version */

#include <signal.h>
#include <fdata.h>
#include <na.h>

extern  DOUBLE fmod();
extern  DOUBLE floor();
extern  DOUBLE ceil();
extern  DOUBLE exp();
extern  DOUBLE log();
extern  DOUBLE sqrt();
extern  DOUBLE sin();
extern  DOUBLE cos();
extern  DOUBLE tan();
extern  DOUBLE asin();
extern  DOUBLE acos();
extern  DOUBLE atan();
extern  DOUBLE sqrt();
extern  DOUBLE atan2();


#define ZILDE(X)        mkobj((X), TYPN, 0, 0)
#define FMOD(X,Y)       fmod((X),(Y))
#define FFLOOR(X)       floor((X))
#define FCEILING(X)     ceil((X))

int     sca_id[]  = {0,     0,   1,  MAX_INT, MIN_INT};
DOUBLE  sca_xid[] = {0.0, 0.0, 1.0, 1.0e200, -1.0e200};
DOUBLE  (*trig_fns[])()={atan, acos, asin, floor, sin, cos, tan};

sca_dya(op, dst, lo, hi)        /* scalar dyadic */
int     op;     /* +,-,*,%, etc  */
MATRIX dst;
MATRIX lo;
MATRIX hi;
{
BYTE   *s  = lo->ADDR;
int     dx = MTYPE(s);
int     nx = MSIZE(s);
BYTE   *t  = hi->ADDR;
int     dy = MTYPE(t);
int     ny = MSIZE(t);
int     dz, qx, qy, n, nc, rc;
MATRIX  p;
int    *ix, *jx, *kx;
DOUBLE *zd, *xs, *ys;
DOUBLE  xr;

       /* null absorbs    X + NULL -> NULL */
if (nx == 0 || ny == 0) return ZILDE(dst);

n  = (nx > ny) ? nx : ny;
nc = (nx > ny) ? MCOLS(s) : MCOLS(t);
qx = nx > 1;
qy = ny > 1;

switch (dx | dy) {
    default: return ERROR;
    case TYPT:  /* character compare */
        if (op == '=' || op == '!') {
           if (mkobj(dst, 'N', n, nc)) return ERROR;
           kx = INT_PTR(dst);
           for(s += HDSIZE, t += HDSIZE; 0 < n--; kx++) {
               *kx = ((*s) == (*t));
               if (op == '!') *kx = ! *kx;
               if (qx) s++;
               if (qy) t++;
               }
           return OK;
           }
        else switch (op) {
        case '>': rc = as_comp(lo, hi);
                  return mkik(dst, rc == 1);
        case '<': rc = as_comp(lo, hi);
                  return mkik(dst, rc == -1);
        case '+': return ds_plus(dst, lo, hi);
        case '-': return ds_minus(dst, lo, hi);
        case '*': return ds_mult(dst, lo, hi);
        case '%':
        case '#': return ds_divmod(op, dst, lo, hi);
        default:  return ERROR;
            break;
            }
        break;
    case (TYPN | TYPT): /* mixed compare */
        return (op == '=' || op == '!') ? mkobj(dst, 'N', n, nc) : ERROR;
    case TYPN:          /* integer */
    case TYPD:  break;  /* double  */
    case TYPN|TYPD:     /* mixed */
        p = (dx == TYPN) ? lo : hi;
        if (fx_cvd(dst, p) || sa_swap(p, dst)) return ERROR;
        dx = dy = TYPD;
        }

dz = (op == '=' || op == '!' || op == '<' ||
      op == '>' || op == '1' || op == '2') ? TYPN : TYPD;
if (((dx | dy) & TYPD) && dz == TYPN)
        rc = mkobj(dst, dz, n, nc);
else    rc = (nx > ny) ? sasa(dst, lo) : sasa(dst, hi);
if (rc) return ERROR;
if (dx == TYPN)         /* N+N->N */
    for(ix = INT_PTR(lo), jx = INT_PTR(hi), kx = INT_PTR(dst);
        0 < n-- && rc == 0; kx++) {
        switch (op) {
        default:
        case '+' : *kx = (*ix)+(*jx);   /* plus */
            break;
        case '-' : *kx = (*ix)-(*jx);   /* minus */
            break;
        case '*' : *kx = (*ix)*(*jx);   /* times */
            break;
        case '%' :  if (rc = (0 == *jx)) break;
                *kx = (*ix)/(*jx);      /* divide */
            break;
        case '#' :  if (rc = (0 == *jx)) break;
                *kx = (*ix)%(*jx);      /* #mod */
                 if (*kx < 0) *kx += *jx;
            break;
        case 'f' : *kx = *ix;            /* floor .. min */
               if ((*ix) > (*jx)) *kx = *jx;
            break;
        case 'c' : *kx = *ix;            /* cieling .. max */
               if ((*ix) < (*jx)) *kx = *jx;
            break;
        case '&' : *kx = (*ix) & (*jx); /* #and */
            break;
        case '|' : *kx = (*ix) | (*jx); /* #or */
            break;
        case '<' : *kx = ((*ix) < (*jx));
            break;
        case '>' : *kx = ((*ix) > (*jx));
            break;
        case '=' : *kx = ((*ix) == (*jx));
            break;
        case '!' : *kx = ((*ix) != (*jx));          /* #ne */
            break;
        case '1' : *kx = ((*ix) <= (*jx));          /* <=  */
            break;
        case '2' : *kx = ((*ix) >= (*jx));          /* >=  */
            break;
            }
    if (qx) ix++;
    if (qy) jx++;
    }
else if (dz == TYPN)    /* D+D->N */
    for(xs = DBL_PTR(lo), ys = DBL_PTR(hi), kx = INT_PTR(dst); 0<n--; kx++) {
        switch (op) {
        default:
        case '&' : /* *kx = (*xs) & (*ys);    #and */
            break;
        case '|' : /* *kx = (*xs) | (*ys);    #or */
            break;
        case '<' : *kx = ((*xs) < (*ys));
            break;
        case '>' : *kx = ((*xs) > (*ys));
            break;
        case '=' : *kx = ((*xs) == (*ys));
            break;
        case '!' : *kx = ((*xs) != (*ys));          /* #ne */
            break;
        case '1' : *kx = ((*xs) <= (*ys));          /* <=  */
            break;
        case '2' : *kx = ((*xs) >= (*ys));          /* >=  */
            break;
            }
    if (qx) xs++;
    if (qy) ys++;
    }
else            /* D+D->D */
    for(xs = DBL_PTR(lo), ys = DBL_PTR(hi), zd = DBL_PTR(dst); 0<n--; zd++) {
        switch (op) {
        default:
        case '+' : *zd = (*xs)+(*ys);   /* plus */
            break;
        case '-' : *zd = (*xs)-(*ys);   /* minus */
            break;
        case '*' : *zd = (*xs)*(*ys);   /* times */
            break;
        case '%' :  if (rc = (0.0 == *ys)) break;
                *zd = (*xs)/(*ys);      /* divide */
            break;
        case '#' :  if (rc = (0.0 == *ys)) break;
                if (*xs < 0.0) *zd = *ys - FMOD(-*xs, *ys);
                else *zd = FMOD (*xs, *ys);     /* #mod */
            break;
        case 'f' : *zd = *xs;           /* floor .. min */
               if ((*xs) > (*ys)) *zd = *ys;
            break;
        case 'c' : *zd = *xs;           /* ceiling .. max */
               if ((*xs) < (*ys)) *zd = *ys;
            break;
            }
    if (qx) xs++;
    if (qy) ys++;
    }
return rc;
}

sca_mon(fc, dst, src)   /* monadic operation */
int     fc;             /* utility operations on integers */
MATRIX  dst;            /* same size */
MATRIX  src;            /* src = integer vector */
{
BYTE   *s  = src->ADDR;
int     n  = MSIZE(s);
int     nc = MCOLS(s);
int     dx = MTYPE(s);
int    *di, dz, k, rc;
DOUBLE *x, xr;
if (n == 0) return sasa(dst, src);      /* no change */
if  (dx == TYPN) dz = dx;
else dz =  (fc == '!' || fc == '~' || fc == '*') ? TYPN : TYPD;
rc = (dx == TYPN || dz == TYPD)
      ? sasa (dst, src)
      : mkobj(dst, dz, n, nc);
if (rc) return rc;
if (dx == TYPN)         /* N->N */
    for (di = INT_PTR(dst); 0 < n--; di++)
        switch(fc) {
        case '-': *di = -*di;   /* negate */
                break;
        case '+':       /* absolute value */
                if (*di < 0) *di = -*di;
                break;
        case '*':       /* signum */
                if (*di > 0) *di = 1;
                else if (*di == 0) *di = 0;
                else *di = -1;
                break;
        case '~':       /* not 1->0 and 0->1 */
                *di = 1 & ~ *di;
                break;
        case '!':       /* logicise */
                *di = (*di != 0);
                break;
        case '?':   /* random throw */
                k = *di;
                *di = (k > 0) ? rand() % k : 0;
        default:        /* copy */
                break;
        }
else if (dz == TYPN)    /* D -> N */
    for (di = INT_PTR(dst), x = DBL_PTR(src); 0 < n--; di++, x++)
        switch(fc) {
        case '*':       /* signum */
                if (*x > 0.0) *di = 1;
                else if (*x == 0.0) *di = 0;
                else *di = -1;
                break;
        case '~':       /* not 1->0 and 0->1 */
                *di = (*x == 0.0) ? 1 : 0;
                break;
        case '!':       /* logicise */
                *di = (*x != 0.0);
                break;
        }
else                    /* D -> D */
    for (x = DBL_PTR(dst); 0 < n-- && rc == 0; x++)
        switch(fc) {
        case '-': *x = -*x;   /* negate */
                break;
        case '+':       /* absolute value */
                if (*x < 0) *x = -*x;
                break;
        case 'f':       /* floor */
                *x = FFLOOR(*x);
                break;
        case 'c':       /* cieling */
                *x = FCEILING(*x);
                break;
        default:        /* copy */
                break;
        }
return rc;
}

sca_sum(fc, dst, src)   /* summation */
int     fc;             /* [A-E][F-I] */
MATRIX  dst;            /* DST <- f / HI  or DST <- f \ HI  */
MATRIX  src;            /* src = vector -> fc/src  matrix: sum rows */
{
BYTE   *s = src->ADDR;
int     n  = MSIZE(s);
int     nc = MCOLS(s);
int     dx = MTYPE(s);
int     j, kc, nr, qs, rc, sn;
int    *ix, *jx;
DOUBLE *xs,  *zd;

fc = fc - 'A';  /* multi thread coding */
kc = fc % 5;    /* function */
qs = fc > 4;    /* operator */

if (n == 0)     /* small min and max values .. ok for province */
        return (qs) ? ZILDE (dst) : mkik(dst, sca_id[kc]);
else if ((dx != TYPN && dx != TYPD) || nc == 0) return ERROR;

nr  = n / nc;   /* rows */
if (nr == 1) {
    nr = nc;
    nc = 1;     /* vector */
    }
if(rc = (qs) ? sasa(dst, src) : mkobj(dst, dx, nc, nc)) return rc;
if (dx == TYPN) {
    ix = INT_PTR(dst);  /* result */
        if (qs) {       /* scan   */
        for (ix += nc, nr--, sn =-1; 0 < nr--; sn = -sn)
            for (j = 0; j < nc; j++, ix++)
            switch (kc) {
            default:
            case 0   : *ix += ix[-nc];
                    break;
            case 1   : *ix = ix[-nc]+sn * *ix;  /* alternating sum */
                    break;
            case 2   : *ix *= ix[-nc];
                    break;
            case 3   : if (ix[-nc] < *ix) *ix = ix[-nc];
                    break;
            case 4   : if (ix[-nc] > *ix) *ix = ix[-nc];
                    break;
            }
        }
        else {          /* start from zero, and accumulate */
        jx = INT_PTR(src);       /* source */
        for (j = 0; j < nc; j++) ix[j] = sca_id[kc];
        for (sn =1; 0 < nr--; sn = -sn)
            for (j = 0; j < nc; j++, jx++)
                switch (kc) {
                default:
                case 0   : ix[j] += *jx;
                        break;
                case 1   : ix[j] += sn * *jx;   /* alternating sum */
                        break;
                case 2   : ix[j] *= *jx;
                        break;
                case 3   : if (ix[j] > *jx) ix[j] = *jx;
                        break;
                case 4   : if (ix[j] < *jx) ix[j] = *jx;
                        break;
                }
        }
    }
else {  /* D->D */
    zd = DBL_PTR(dst);  /* result */
    if (qs) {           /* scan   */
        for (zd += nc, nr--, sn =-1; 0 < nr--; sn = -sn)
            for (j = 0; j < nc; j++, zd++)
            switch (kc) {
            default:
            case 0   : *zd += zd[-nc];
                    break;
            case 1   : *zd = zd[-nc]+sn * *zd;  /* alternating sum */
                    break;
            case 2   : *zd *= zd[-nc];
                    break;
            case 3   : if (zd[-nc] < *zd) *zd = zd[-nc];
                    break;
            case 4   : if (zd[-nc] > *zd) *zd = zd[-nc];
                    break;
            }
        }
        else {          /* start from zero, and accumulate */
        xs = DBL_PTR(src);       /* source */
        for (j = 0; j < nc; j++) zd[j] = sca_xid[kc];
        for (sn =1; 0 < nr--; sn = -sn)
            for (j = 0; j < nc; j++, xs++)
                switch (kc) {
                default:
                case 0   : zd[j] += *xs;
                        break;
                case 1   : zd[j] += sn * *xs;   /* alternating sum */
                        break;
                case 2   : zd[j] *= *xs;
                        break;
                case 3   : if (zd[j] > *xs) zd[j] = *xs;
                        break;
                case 4   : if (zd[j] < *xs) zd[j] = *xs;
                        break;
                }
        }
    }
return OK;
}

fx_math(dst, src)       /* math utility functions */
MATRIX  dst;            /* always double */
MATRIX  src;
{
DOUBLE (*zf)();
int     fc = MELSIZ(dst->ADDR);
BYTE   *s  = src->ADDR;
int     n  = MSIZE(s);
int     nc = MCOLS(s);
int     dx = MTYPE(s);
int     rc = 0;
DOUBLE *x, *xr;

if (fc == 'S') fc = MINDX(dst->ADDR);
if (n == 0) return sasa(dst, src);      /* no change */
if (dx == TYPN) {
    if (fx_cvd(dst, src) || sa_swap(src, dst)) return fc;
    }
if (fc == F_ATAN && 0 == (nc & 1)) {    /* for complex numbers */
    rc = mkobj(dst, TYPD, n=n/2, nc/2);
    signal(SIGFPE, SIG_IGN);
    for (xr = DBL_PTR(dst), x = DBL_PTR(src); rc == 0 && 0 < n--; xr++, x+=2)
        *xr = atan2(*x, *(x+1));
    return rc;
    }

rc = mkobj(dst, TYPD, n, nc);
if      (fc == F_EXP ) zf = exp;
else if (fc == F_LN)   zf = log;
else if (fc == F_SQRT) zf = sqrt;
else return ERROR;
signal(SIGFPE, SIG_IGN);
for (xr = DBL_PTR(dst), x = DBL_PTR(src); rc == 0 && 0 < n--; xr++, x++)
    *xr = (*zf)(*x);
return rc;
}

fx_circle(dst, f, src) /* math utility functions (trig) */
MATRIX  dst;            /* always double */
MATRIX  f;
MATRIX  src;
{
int     nf = MSIZE(f->ADDR);
BYTE   *s  = src->ADDR;
int     n  = MSIZE(s);
int     nc = MCOLS(s);
int     rc = 0, k, nmax;
int    *fc = INT_PTR(f);
int     df = 1;
int     dx = 1;
DOUBLE *x, *xr;

if (n == 0 || nf == 0) return ZILDE(dst);  /* empty object */
nmax = (nf > n) ? nf : n;
if (nf != n) {
    if ((nf * n) != nmax) return ERROR;
    if (nf < n) df = 0;
    else {
        dx = 0;
        nc = MCOLS(f->ADDR);
        }
    }

signal(SIGFPE, SIG_IGN);        /* ignore domain error for time being */

if (MTYPE(s) == TYPN) {
    if (fx_cvd(dst, src) || sa_swap(src, dst)) return ERROR;
    }
rc = mkobj(dst, TYPD, nmax, nc);
xr = DBL_PTR(dst);
x =  DBL_PTR(src);
for ( ; rc == 0 && 0 < nmax--; xr++) {
    k = *fc;
    if (k < -3 || k > 3) rc = k;
    else  *xr = (trig_fns[k+3])(*x);
    if (df) fc++;
    if (dx) x++;
    }
return rc;
}

mk_ck(MATRIX dst, int c)        /* character 'scalar' */
{
if (mkobj(dst, TYPT, 1, 1)) return ERROR;
*(TEXT_PTR(dst)) = c;
return OK;
}

ss_strip_bits(BYTE *si, int n)  /* strip bits from ASCII digits */
{
for (; 0 < n--; si++) *si &= 0X0F;
return OK;
}

ds_trim(MATRIX dst, MATRIX src) /* delete leading zeros */
{
BYTE   *si = TEXT_PTR(src);
int     nx = MSIZE(src->ADDR);
BYTE   *di;
int     cnt, i, n;

for (cnt = 0, i= nx; 0 < i--; si++) {
    if (*si & 0X0F) break;
    else cnt++;
    }

n = nx - cnt;
if (n == 0) return mk_ck(dst, '0');

if (mkobj(dst, TYPT, n, n)) return ERROR;
for (di =TEXT_PTR(dst), si = TEXT_PTR(src) + cnt; 0 < n--; di++, si++)
    *di = *si | 0X30;

return OK;
}

ds_prefix(MATRIX dst, MATRIX src, int prefix)    /* normalise dst */
{
BYTE   *si = TEXT_PTR(src);
int     nx = MSIZE(src->ADDR);
int     dx = MINDX(src->ADDR);
BYTE   *di;
int     n;

if (nx == dx) return mk_ck(dst, '0');

n = nx - dx + (prefix > 0);

if (mkobj(dst, TYPT, n, n)) return ERROR;
di = TEXT_PTR(dst);

if (prefix) {
    *di++ = prefix;
    n--;
    }

for(si =TEXT_PTR(src) + dx; 0 < n--; di++, si++) *di = 0X30 | *si;
return OK;
}

ds_plus(MATRIX dst, MATRIX lo, MATRIX hi)       /* DST<-LO+HI as strings */
{
MATRIX  px;
BYTE   *si = TEXT_PTR(lo);
BYTE   *zi = TEXT_PTR(hi);
int     nx = MSIZE(lo->ADDR);
int     ny = MSIZE(hi->ADDR);
BYTE   *di;
int     sx = 1, sy = 1;
int     cc, k, max, n, x;

if (nx == 0 || ny == 0) return mkobj (dst, TYPT, 0, 0);

if (nx > 0 && *si == '-') {
    MINDX(lo->ADDR) = 1;
    *si = 0;
    sx  = -1;
    si++;
    nx--;
    }

if (ny >0 && *zi == '-') {
    MINDX(hi->ADDR) = 1;
    *zi = 0;
    sy = -1;
    zi++;
    ny--;
    }


if (sx * sy > 0) {      /* same sign */
    cc = (nx > ny)
        ? ss_add(di = si, max = nx, zi, ny)
        : ss_add(di = zi, max = ny, si, nx);
    if (sx > 0 && cc == 0)      /* same string length */
    return (di == si) ? sasa(dst, lo) : sasa(dst, hi);
    n = max + (sx < 0) + (cc > 0);
    if (mkobj (dst, TYPT, n, n)) return ERROR;
    zi = TEXT_PTR(dst);
    if (sx < 0) *zi++ = '-';
    if (cc)     *zi++ = '0' + cc;
    memcpy(zi, di, max);
    return OK;
    }

if (nx > ny) k = 1;             /* different signs, so subtract */
else if (nx < ny) k = -1;
else if (nx == ny) k = memcmp(si, zi, nx);

if (k == 0) return mk_ck(dst, '0');

ss_strip_bits(zi, ny);
ss_strip_bits(si, nx);

if (k > 0) ds_decz(px = lo, zi, ny, 0);
else       ds_decz(px = hi, si, nx, 0);

k = ((sx < 0 && k > 0) || (sx > 0 && k < 0)) ? '-' : 0;

return ds_prefix(dst, px, k);
}

int     ss_add(BYTE *px, int nx, BYTE *py, int ny)      /* px += py */
{
int     x, cc;

for (px += (nx-1), py += (ny-1), cc = 0 ; 0 < nx--; px--) {
    x = cc + (*px & 0X0F);
    if (0 < ny--) x += (0X0F & *py--);
    cc = (x > 9);       /* carry, base 10 */
    x = (cc) ? x-10 : x;
    *px =  x + '0';
    }
return  cc;
}

int as_comp(MATRIX lo, MATRIX hi)       /* test lo > hi */
{
extern  int     memcmp();
int     nx = MSIZE(lo->ADDR);
int     ny = MSIZE(hi->ADDR);
BYTE   *si = TEXT_PTR(lo);
BYTE   *zi = TEXT_PTR(hi);
int     k;

if (nx > ny)      k =  1;       /* compute sign lo-hi */
else if (nx < ny) k = -1;
else {
    k = memcmp(si, zi, nx);
    if (k > 0) k = 1;
    else if (k < 0) k = -1;
    }
return k;
}

ds_minus(MATRIX dst, MATRIX lo, MATRIX hi)      /* DST<- LO-HI */
{
int     nx = MSIZE(lo->ADDR);
int     ny = MSIZE(hi->ADDR);
BYTE   *si = TEXT_PTR(lo);
BYTE   *zi = TEXT_PTR(hi);
BYTE   *di;
int     cc, k, max, n, sx = 1, sy = 1;

if (nx == 0 || ny == 0) return mkobj (dst, TYPT, 0, 0);

if (nx && *si == '-') {         /* LO negative */
    sx = -1;
    *si = 0;
    si++;
    nx--;
    }

if (ny && *zi == '-') {         /* HI negative */
   *zi = 0;
    sy = -1;
    *zi++;
    ny--;
    }

if (sx == sy) {
    ss_strip_bits(TEXT_PTR(lo), nx);
    ss_strip_bits(TEXT_PTR(hi), ny);

    if (nx > ny)      k = 1;
    else if (nx < ny) k = -1;
    else              k = memcmp(si, zi, nx);

    if (k == 0) return mk_ck(dst, '0');
    else if (k > 0) {
        ds_decz(lo, zi, ny, 0);
        return ds_prefix(dst, lo, (sx > 0) ? 0 : '-');
        }
    else if (k < 0) {
        ds_decz(hi, si, nx, 0);
        return ds_prefix(dst, hi, (sx > 0) ? '-' : 0);
        }
    }

cc = (nx > ny)  /* do an addition */
    ? ss_add(di = si, max = nx, zi, ny)
    : ss_add(di = zi, max = ny, si, nx);

n =  max + (sx < 0) + (cc > 0);
if (mkobj(dst, TYPT, n, n)) return ERROR;
zi = TEXT_PTR(dst);
if (sx < 0) *zi++ = '-';
if (cc)     *zi++ = '0' + cc;
memcpy(zi, di, max);
return OK;
}

ds_mult(MATRIX dst, MATRIX lo, MATRIX hi)       /* multiply */
{
BYTE   *si = TEXT_PTR(lo);
BYTE   *zi = TEXT_PTR(hi);
int     nx = MSIZE(lo->ADDR);
int     ny = MSIZE(hi->ADDR);
BYTE   *di;
int     n_sign = 1;
int     n, nxy, j, k, x, cc;

if (*si == '-') {       /* sign of lo */
    n_sign = -n_sign;
    si++;
    nx--;
    }

if (*zi == '-') {       /* sign of hi */
    n_sign = -n_sign;
    zi++;
    ny--;
    }

if (nx == 1 && *si == '0' || ny == 1 && *zi == '0') return mk_ck(dst, '0');
else if (nx * ny == 0) return mkobj(dst, TYPT, 0, 0);

ss_strip_bits(si, nx);
ss_strip_bits(zi, ny);

n = nxy = nx + ny;
if (n_sign < 0) n++;
if (mkobj(dst, TYPT, n, n)) return ERROR;

di = TEXT_PTR(dst) + n-1;
si += (nx-1);
zi += (ny-1);

for (k = cc = 0; k < nxy-1; k++) {
    for (x = cc, j = 0; j <= k; j++) {
        if (j < nx && k-j < ny ) x += si[-j] * zi[-(k-j)];
        }
    cc = x/10;
    *di-- = '0' + (x % 10);
    }

*di = '0' + cc;
if (cc) {
    if (n_sign < 0) *(TEXT_PTR(dst)) = '-';
    return OK;
    }

MINDX(dst->ADDR) = 1 +(n_sign < 0);     /* extra digit ? */
return ds_prefix(lo, dst, (n_sign < 0) ? '-' : 0) ||
       sasm (dst, lo);
}

ds_divmod(int op, MATRIX dst, MATRIX lo, MATRIX hi)     /* Z<-LO op HI */
{
extern  MATRIX  pm_sym;
MATRIX  sys_nlink();
MATRIX  ds_t9();
MATRIX  x9 = NULL;      /* multiplication table */
MATRIX  xr = NULL;      /* remainder */
MATRIX  xq = NULL;      /* quotient */
int     nx = MSIZE(lo->ADDR);
int     ny = MSIZE(hi->ADDR);
BYTE   *di, *si, *t9, *zq;
int     dx, i, ip, k, kz, n, nc9, nt, nq;

if (nx == 0 || ny == 0)
    return mkobj (dst, TYPT, 0, 0);

if (op == '%') {
    xq = dst;
    xr = sys_nlink(pm_sym, "$REMAINDER", NULL, 0);
    }
else if (op == '#') {
    xq = sys_nlink(pm_sym, "$QUOTIENT", NULL, 0);
    xr = dst;
    }
else return ERROR;

k = as_comp(lo, hi);

if (k == 0)     return mk_ck(xr, '0') || mk_ck(xq, '1');
else if (k < 0) return sasa (xr, lo)  || mk_ck(xq, '0');

if (ny == 1) {          /* trivial case. divide by 0 or 1 */
    si = TEXT_PTR(hi);
    if (*si == '0') return ERROR;
    else if (*si == '1') return mk_ck(xr, '0') || sasa(xq, lo);
    else return ds_idiv(xq, lo, *si, &k) || mk_ck(xr, k);
    }

/* really difficult stuff */

ss_strip_bits(TEXT_PTR(lo), nx);
ss_strip_bits(TEXT_PTR(hi), ny);

x9  = ds_t9("$QR_T9", hi);
t9  = TEXT_PTR(x9);
nc9 = MCOLS(x9->ADDR);
nq  = nx-ny+1;

if (mkobj(xq, TYPT, nq, nq)) return ERROR;
zq = TEXT_PTR(xq);
for (i = 0; i < nq; i++) zq[i] = '0';

for (di = TEXT_PTR(lo);;) {
    si = TEXT_PTR(hi);
    dx = MINDX(lo->ADDR);
    kz = (nx - dx) - ny;
    if (kz < 0) break;          /* test hi < lo */
    else if (kz == 0 && 0 > memcmp(di + dx, si, ny)) break;
    kz -= (0 < memcmp(si, di + dx, ny));        /* decimal place */

    for (ip = 9, i = 0; i < 9; i++) {   /* get the correct digit */
        si = t9 + i*nc9;
        nt = nc9;
        if (*si == 0) {
            si++;
            nt--;
            }
        if ((nx-dx == kz+nt && 0 < memcmp(si, di + dx, nt)) ||
             nx-dx <  kz+nt) {
            ip = i;
            break;
            }
        }

    si = t9 + (ip-1)*nc9;
    nt = nc9;
    if (*si == 0) {
        si++;
        nt--;
        }
    zq[nq-(kz+1)] = '0'+ip;
    ds_decz(lo, si, nt, kz);
    }

k = MINDX(lo->ADDR);
n = nx - k;
if (n == 0)  mk_ck(xr, '0');
else {
     if (mkobj(xr, TYPT, n, n)) return ERROR;
     for (si = TEXT_PTR(xr), di += k; 0 < n--; si++, di++) *si = 0X30 | *di;
     }

return sa_swap(lo, xq) || ds_trim(xq, lo);
}

/* dst<- dst-(10**nz)*(s, ns) */

ds_decz(MATRIX dst, BYTE *src, int n, int nz)  /* compute in situ */
{
int     nx = MSIZE(dst->ADDR);
int     dx = MINDX(dst->ADDR);
BYTE   *si, *di;
int     bb, j, x;

nx = nx - dx;                                   /* precision of dst */
di = TEXT_PTR(dst) + dx + (nx - 1) - nz;        /* skip nz zeros */
si = src + n - 1;

for (nx -= nz, bb = 0; 0 < nx--; di--) {
    x = (*di) - bb;
    if (0 < n--) x -= (*si--);
    if (x < 0) {
        bb = 1;
        x += 10;
        }
    else bb = 0;
    *di = x;
    }

/*      count leading zeros */

for (j = 0, di = TEXT_PTR(dst), nx = MSIZE(dst->ADDR); 0 < nx--; di++) {
    if (*di) break;
    else j++;
    }

MINDX(dst->ADDR) = j;
return OK;
}

MATRIX  ds_t9 (BYTE *name, MATRIX src)  /* multiplication table */
{
extern  MATRIX  pm_sym;
MATRIX  sys_nlink();
MATRIX  p;
int     nx = MSIZE(src->ADDR);
BYTE   *si, *di, *zi;
int     cc, j, k, n, x;

p = sys_nlink(pm_sym, name, NULL, 0);
n = nx+1;
if (p == NULL || mkobj(p, 'T', 9*n, n))  return NULL;
si = TEXT_PTR(src);
di = TEXT_PTR(p);
memcpy(di+1, si, nx);
ss_strip_bits(di, n);
for (k = 2, di += n; k < 10; k++, di += n) {    /* src * 'k' */
    zi = di + n  - 1;
    si = TEXT_PTR(src) + nx - 1;
    for (cc = j = 0; j < nx; j++, si--, zi--) {
        x = cc + k * (*si & 0X0F);
        *zi =  x % 10;
        cc = x / 10;
        }
    *zi = cc;
    }
return p;
}

ds_idiv(MATRIX dst, MATRIX src, int k, int *r)  /* dst <- src % val */
{
BYTE   *si = TEXT_PTR(src);
int     nx = MSIZE(src->ADDR);
int     x  = si[0] & 0X0F;
BYTE   *di;
int     cc, n;

k = k & 0X0F;
n = nx - (k > x);
if (mkobj(dst, TYPT, n, n)) return ERROR;

di = TEXT_PTR(dst);

if (k > x) {
    cc = x;
    si++;
    nx--;
    }
else    cc = 0;

for (; 0 < nx--; di++, si++) {
    x = 10*cc + (*si & 0X0F);
    *di = '0' + x / k;
    cc  = x % k;
    }
*r = '0' + cc;          /* remainder */
return OK;
}


